#!/usr/bin/perl
#-------------------------------------------------------------------------
# Automatic Csound Score Generation by Hans Mikelson January 1999
#
# This program generates a fractal csound score for use in granular synthesis.
# In Perl a '#' indicates the rest of the line is a comment.
# All scalar variables begin with a '$'.
#-------------------------------------------------------------------------
# Main Program begins here

# Open the score file for writing and give it a table for the oscilator
# To print a line of text to the score use print SCORE "text\n";
$scorefile = "fscore.sco";
open (SCORE, ">$scorefile");
print SCORE "; Sine Table\n";
print SCORE "f1 0 8192 10 1\n\n";

# Significant figures printed out is 3 since $sf=1000
$sf = 1000;

# GRAIN CLOUD GENERATOR
# The following lines call the subroutine gs1 which
# then generates a cloud based on the parameters given.
# Start time, duration, amplitude, frequency and number of grains
# are specified.  A minimum value and a multiplier are provided.
#    Sta        Dur       Amp          Fqc         Grains
#    Min   Rng  Min  Rng  Min    Rng   Min   Rng   #
#gs1(0,    2,   .1,  4,   10,    2000, 200,  3000, 100);

# STREAM GENERATOR
# The following lines call the subroutine st1 which generates
# a string of notes one after the other.
#   Sta   Dur       Amp           Fqc          Fco         Rez     Notes
#         Min  Rng  Min    Rng    Min   Rng    Min  Rng    Min Rng
st2(0,    .1,  4,   20000, 20000, 200,  200,   400, 20000, 0, .4,  60);

close SCORE;

# Now it's done.  The following lines are helpful when rendering on an SGI.
# perl fscore2.pl
# cs fscore.orc fscore.sco
# soundeditor test
# nedit fscore.sco &

#-------------------------------------------------------------------------
# SUBROUTINES
# $_[i] is a special variable which returns the ith parameter in a
# subroutine call.  All variables are global unless you use the word my in
# front of them.
#-------------------------------------------------------------------------

# Mandeloop
# This is the main loop for calculating the Mandelbrot set.
sub mandeloop
 {
  my $cr = $_[0];
  my $ci = $_[1];
  my $zr = 0;
  my $zi = 0;
  my $zzr = 0;

# The {} means this is a hash or an associative array
# It sort of means that it is referenced by a string rather than by numbers.
# The third parameter $_[2] when this function is called is a string.
  $grain{$_[2]} = 0;

# %grain is a global hash which I use to store the score event variables
# In this case I use it to count the iterations of the Mandelbrot inner loop.
  while (($grain{$_[2]} < 200) && ($zr*$zr + $zi*$zi < 4))
   {
#   z = z^2 + c ; Where z & c are complex numbers.
    $zzr = $zr*$zr - $zi*$zi + $cr;
    $zi  = 2*$zr*$zi + $ci;
    $zr  = $zzr;
    $grain{$_[2]}++;
   }
 }

# Barnsleyj1
# This one was not meant to be used this way but...
# This is another fractal similar to the mandelbrot set.
sub barnsleyj1
 {
  my $cr = $_[0];
  my $ci = $_[1];
  my $zzr = 0;

# z(0) = pixel;
  $grain{$_[4]} = $_[2];
  $grain{$_[5]} = $_[3];

# if real(z) >= 0
#   z(n+1) = (z-1)*c
# else
#   z(n+1) = (z+1)*c

  if ($grain{$_[4]} >= 0)
   { 
    $zzr = ($grain{$_[4]}-1)*$cr - $grain{$_[5]}*$ci;
    $grain{$_[5]} = $grain{$_[5]}*$cr + $ci*($grain{$_[4]} - 1);
    $grain{$_[4]} = $zzr;
   }
  else
   { 
    $zzr = ($grain{$_[4]} + 1)*$cr - $grain{$_[5]}*$ci;
    $grain{$_[5]} = $grain{$_[5]}*$cr + $ci*($grain{$_[4]} + 1);
    $grain{$_[4]} = $zzr;
   }
 }

# Henon Attractor
# $grain{$_[3]} is x and $grain{$_[4]} is y.
sub henon
 {
  my $a = $_[0];
  my $b = $_[1];
  my $xx = 1 + $grain{$_[3]} - $a*$grain{$_[2]}*$grain{$_[2]};
  my $yy = $b*$grain{$_[2]};
  $grain{$_[2]} = $xx;
  $grain{$_[3]} = $yy;
 }

# Circle squared
sub circle
 {
  my $x = $_[0];
  my $y = $_[1];
  my $s = $_[2];
  my $m = $_[3];
  my $c = ($s*($x^2 + $y^2)) % $m;
  $grain{$_[4]} = $c;
 }

# Hopalong
sub hopalong
 {
  my $x = $grain{$_[2]};
  my $y = $grain{$_[3]};
  
 }

# This routine calls several of the above routines to generate
# granular cloud based on fractals.  This routine is a good one to
# copy and modify to try to get some different sounds.
sub gs1
 {
  $stao = $_[0]; # Start minimum or offset
  $stam = $_[1]; # Start range
  $duro = $_[2]; # Duration min
  $durm = $_[3]; # Duration range
  $ampo = $_[4]; # Amplitude min
  $ampm = $_[5]; # Amplitude range
  $fqco = $_[6]; # Frequency min
  $fqcm = $_[7]; # Frequency range
  $iter = $_[8]; # Number of iterations

  $maxpan = 10;  # Starting maximum for panning.
# Panning is normalized from 0-1 based on the maximum value.

# Loop for the number of grains generated.
  for ($i=0; $i<$iter; $i++)
   {

#   The steps $sx & $sy are used in functions like the mandelbrot set
#   which rely on x & y coordinates.
    $sy = int($i/sqrt($iter));
    $sx = $i % int(sqrt($iter));

#   Use henon to determine the start and duration times.
#   This would be a good place to modify some coefficients or which fractal
#   is called for which score event variable.
    henon(1.4, .3, 'sta', 'dur');
    barnsleyj1(1.2, .5, -1.3+.1*$sx, .25+.1*$sy, 'amp', 'fqc');
    mandeloop(-1.5+.01*$sx, -1+.01*$sy, 'pan');

#   Normalize the pan between 0 & 1.
    if (abs($grain{'pan'})>$maxpan) {$maxpan = abs($grain{'pan'})}

#   Perform the appropriate offsets and multiplications on the data
    $sta = $stam*int($sf*abs($grain{'sta'}))/$sf + $stao;
    $dur = $durm*int($sf*abs($grain{'dur'}))/$sf + $duro;
    $amp = $ampm*int($sf*abs($grain{'amp'}))/$sf + $ampo;
    $fqc = $fqcm*int($sf*abs($grain{'fqc'}))/$sf + $fqco;
    $pan = int($sf*abs($grain{'pan'}))/$sf/$maxpan;

#   Everything is ready so print the score event
    print SCORE "i1 $sta $dur $amp $fqc $pan\n";
   }
}

# An earlier grain generator.  See the one above for comments.
sub gs2
 {
  $stao = $_[0];
  $stam = $_[1];
  $duro = $_[2];
  $durm = $_[3];
  $ampo = $_[4];
  $ampm = $_[5];
  $fqco = $_[6];
  $fqcm = $_[7];
  $iter = $_[8];

  $maxpan = .001;

  for ($i=0; $i<$iter; $i++)
   {
    henon(1.4, .3, 'sta', 'dur');
    henon(1.4, .3, 'amp', 'fqc');
    henon(1.4, .3, 'pan', 'atk');
    if (abs($grain{'pan'})>$maxpan) {$maxpan = abs($grain{'pan'})}
    
    $sta = $stam*int($sf*abs($grain{'sta'}))/$sf + $stao;
    $dur = $durm*int($sf*abs($grain{'dur'}))/$sf + $duro;
    $amp = $ampm*int($sf*abs($grain{'amp'}))/$sf + $ampo;
    $fqc = $fqcm*int($sf*abs($grain{'fqc'}))/$sf + $fqco;
    $pan = int($sf*abs($grain{'pan'}))/$sf/$maxpan;

    print SCORE "i1 $sta $dur $amp $fqc $pan\n";
   }
}

# The following subroutine generates a stream of notes based on the fractals.
sub st1
 {
  $stao = $_[0];
  $duro = $_[1];
  $durm = $_[2];
  $ampo = $_[3];
  $ampm = $_[4];
  $fqco = $_[5];
  $fqcm = $_[6];
  $iter = $_[7];

  $maxpan = 2;

  $sta = $stao;
  $dur = $duro;

  for ($i=0; $i<$iter; $i++)
   {
    henon(1.4, .3, 'amp', 'fqc');
    henon(1.4, .3, 'pan', 'atk');
    if (abs($grain{'pan'})>$maxpan) {$maxpan = abs($grain{'pan'})}


#   In this case all durations are constant and each note follows another.
    $amp = $ampm*int($sf*abs($grain{'amp'}))/$sf + $ampo;
    $fqc = $fqcm*int($sf*abs($grain{'fqc'}))/$sf + $fqco;
    $pan = int($sf*abs($grain{'pan'}))/$sf/$maxpan;

    print SCORE "i2 $sta $dur $amp $fqc $pan\n";
    $sta = $sta + $dur;
   }
}

# The following subroutine generates a stream of notes based on the fractals.
sub st2
 {
  $stao = $_[0];
  $duro = $_[1];
  $durm = $_[2];
  $ampo = $_[3];
  $ampm = $_[4];
  $fqco = $_[5];
  $fqcm = $_[6];
  $fcoo = $_[7];
  $fcom = $_[8];
  $rezo = $_[9];
  $rezm = $_[10];
  $iter = $_[11];

  $maxpan = 2;

  $sta = $stao;
  $dur = $duro;

  for ($i=0; $i<$iter; $i++)
   {
    henon(1.4, .3, 'amp', 'fqc');
    henon(1.4, .3, 'pan', 'fco');
    $sy = int($i/sqrt($iter));
    $sx = $i % int(sqrt($iter));

    mandeloop(-1.5+.01*$sx, -1+.01*$sy, 'rez');
    if (abs($grain{'pan'})>$maxpan) {$maxpan = abs($grain{'pan'})}

#   In this case all durations are constant and each note follows another.
    $amp = $ampm*int($sf*abs($grain{'amp'}))/$sf + $ampo;
    $fqc = $fqcm*int($sf*abs($grain{'fqc'}))/$sf + $fqco;
    $pan = int($sf*abs($grain{'pan'}))/$sf/$maxpan;
    $fco = $fcom*int($sf*abs($grain{'fco'}))/$sf + $fcoo;
    $rez = $rezm*int($sf*abs($grain{'rez'}))/$sf + $rezo;

    print SCORE "i3 $sta $dur $amp $fqc $pan $fco $rez\n";
    $sta = $sta + $dur;
   }
}

# The following subroutine generates a stream of notes based on the fractals.
sub st3
 {
  $stao = $_[0];
  $duro = $_[1];
  $durm = $_[2];
  $ampo = $_[3];
  $ampm = $_[4];
  $fqco = $_[5];
  $fqcm = $_[6];
  $iter = $_[7];

  $maxpan = 2;

  $sta = $stao;
  $dur = $duro;

  for ($i=0; $i<$iter; $i++)
   {
    henon(1.4, .3, 'amp', 'fqc');
    henon(1.4, .3, 'pan', 'atk');
    if (abs($grain{'pan'})>$maxpan) {$maxpan = abs($grain{'pan'})}


#   In this case all durations are constant and each note follows another.
    $amp = $ampm*int($sf*abs($grain{'amp'}))/$sf + $ampo;
    $fqc = $fqcm*int($sf*abs($grain{'fqc'}))/$sf + $fqco;
    $pan = int($sf*abs($grain{'pan'}))/$sf/$maxpan;

    print SCORE "i3 $sta $dur $amp $fqc $pan\n";
    $sta = $sta + $dur;
   }
}

